#!/usr/bin/python
# coding=UTF-8

import subprocess
import re
import itertools
import fitz

""" pdftotext -layout -nopgbrk p1220.pdf - """

def strip_values(items):
    expr_non_alphanum = re.compile(r'[^\w\s]*', re.MULTILINE)
    return [expr_non_alphanum.sub(x, '').strip().replace('\n', ' ') for x in items if x]

class PDFRecordFinder(object):
    field_range_expr = re.compile(r'^(\d+)[-]?(\d*)$')

    def __init__(self, src):
        self.document = fitz.open(src)

    def find_record_table_ranges(self):
        matches = []
        for (page_number, page) in enumerate(self.document):
            header_rects = page.search_for("Record Name:")
            for header_match_rect in header_rects:
                header_match_rect.x0 = header_match_rect.x1  # Start after match of "Record Name: "
                header_match_rect.x1 = page.bound().x1  # Extend to right side of page
                header_text = page.get_textbox(header_match_rect)
                record_name = re.sub(r'[^\w\s\n]*', '', header_text).strip()
                matches.append((record_name, {
                    'page': page_number,
                    'y': header_match_rect.y1 - 5,  # Back up a hair to include header more reliably
                }))
        return matches

    def find_records(self):
        record_ranges = self.find_record_table_ranges()
        for record_index, (record_name, record_details) in enumerate(record_ranges):
            current_rows = []
            next_index = record_index+1
            (_, next_record_details) = record_ranges[next_index] if next_index < len(record_ranges) else (None, {'page': self.document.page_count-1})
            for page_number in range(record_details['page'], next_record_details['page']):
                page = self.document[page_number]
                table_search_rect = page.bound()
                if page_number == record_details['page']:
                    table_search_rect.y0 = record_details['y']
                tables = page.find_tables(
                        clip = table_search_rect,
                        min_words_horizontal = 1,
                        min_words_vertical = 1,
                        horizontal_strategy = "lines_strict",
                        intersection_tolerance = 1,
                        )
                for table in tables:
                    if table.col_count == 4:
                        table = table.extract()
                        # Parse field position (sometimes a cell has multiple
                        # values because IRS employees apparently smoke crack
                        for row in table:
                            first_column_lines = row[0].strip().split('\n')
                            if len(first_column_lines) > 1:
                                for sub_row in self.split_row(row):
                                    current_rows.append(strip_values(sub_row))
                            else:
                                current_rows.append(strip_values(row))
            consecutive_rows = self.filter_nonconsecutive_rows(current_rows)
            yield(record_name, consecutive_rows)

    def split_row(self, row):
        if not row[1]:
            return []
        split_rows = list(itertools.zip_longest(*[x.strip().split('\n') for x in row[:3]], fillvalue=None))
        description = strip_values([row[3]])[0]
        rows = []
        for row in split_rows:
            if len(row) < 3 or not row[2]:
                row = self.infer_field_length(row)
            rows.append([*row, description])
        return rows

    def infer_field_length(self, row):
        matches = PDFRecordFinder.field_range_expr.match(row[0])
        if not matches:
            return row
        (start, end) = ([int(x) for x in list(matches.groups()) if x] + [None])[:2]
        length = str(end-start+1) if end and start else '1'
        return (*row[:2], length)

    def filter_nonconsecutive_rows(self, rows):
        consecutive_rows = []
        last_position = 0
        for row in rows:
            matches = PDFRecordFinder.field_range_expr.match(row[0])
            if not matches:
                continue
            (start, end) = ([int(x) for x in list(matches.groups()) if x] + [None])[:2]
            if start != last_position + 1:
                continue
            last_position = end if end else start
            consecutive_rows.append(row)
        return consecutive_rows

    def records(self):
        return self.find_records()
